#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#ifndef NAN
#define NAN (0.0/0.0)
#endif


#include "dataset.h"
#include "kdtree.h"
#include "range_tree.h"
#include "htable.h"
#include "eclat_items.h"

#define MAXLINE 10000

/////////////////////////////////////////////////
// Set the counting data structure/method here
/////////////////////////////////////////////////


//#define USE_BRUTE_FORCE
//#define USE_KDTREE
#define USE_RANGE_TREE
//#define USE_EXPLICIT_PAIRS		/* create all pairs explicitely, then use eclat */


/////////////////////////////////////////////////
// Set the support measure to use here
/////////////////////////////////////////////////

//#define SUPP_RHO
//#define SUPP_F
#define SUPP_TAU


char linebuf[MAXLINE];
size_t N;
dataset * orig_data;
dataset * whole_data_lrank;
dataset * whole_data_hrank;

size_t
count_items(char * line, size_t * n_categ_items)
{
    char * lcopy = strdup(line);
    size_t n = 0;
    char * str = lcopy;
    char * str2;

    *n_categ_items = 0;
    str2 = strsep(&str, ",;");
    while(str2 != NULL)
    {
	if(isalpha(str2[0]))
	{
	    (*n_categ_items)++;
	}
	n++;
	str2 = strsep(&str, ",;");
    }

    free(lcopy);
    return n;
}

void
print_data(dataset * d)
{
    size_t i, j;

    for(i = 0; i < d->n; i++)
    {
	printf("%f", d->data[i][0]);
	for(j = 1; j < d->n_items; j++)
	{
	    printf(",%f", d->data[i][j]);
	}
	printf("\n");
    }
}



#ifndef USE_RANGE_TREE
void
eclat(size_t i, size_t n_items, dataset * dgreater, dataset * dless,
      size_t minsup, size_t * root, size_t lenroot, htable * freq,
      size_t * mark_gt, size_t * mark_lt)
#endif
#ifdef USE_RANGE_TREE
void
eclat(size_t i, size_t n_items, dataset * dgreater, dataset * dless,
      size_t minsup, size_t * root, size_t lenroot, htable * freq,
      range_tree * rt, index_entry ** indices)
#endif
{
    size_t supp = 0;
    size_t u_bound = 1UL<<30;	/* upper bound on support_tau */
    size_t l_bound = 0;		/* lower bound on support_tau */
    double F = 0, rho = 0;	/* other definitions of support */
    size_t r, s, t, u;
    dataset * d_gt_new = dgreater; 		/* pruned dataset */
    dataset * d_lt_new = dless; 		/* pruned dataset */
#ifdef USE_KDTREE
    KDtree kdtree_less;
    KDtree kdtree_gt;
#endif
    size_t n_lt, n_gt;
    size_t old_n_lt;
    static size_t prune_buf[5000];
    static double bounds[5000];

    //print 'cand', root, len(Dcur)
    //for(s = 0; s < lenroot; s++)
    //	printf("%d,", root[s]);
     //printf("%d\n", dgreater->n);
    if(i > n_items)
	return;

    if(lenroot >= 1)
    {
        // pruning
	if(lenroot >= 2)
	{
	    prune_buf[0] = lenroot - 1;
	    for(r = 0; r < lenroot; r++)
	    {
		t = 1;
		for(s = 0; s < lenroot; s++)
		    if(s != r)
			prune_buf[t++] = root[s];
		if(htable_get(freq, prune_buf) == NULL)
		    return;
	    }
	}

#ifdef SUPP_TAU
	/* create pruned datasets */
#ifdef USE_BRUTE_FORCE
	d_gt_new = dataset_new();
	d_lt_new = dless;
#endif
#ifdef USE_RANGE_TREE
	d_gt_new = dataset_new();
	d_lt_new = dless;
#endif
#ifdef USE_KDTREE
	d_gt_new = dataset_new();
	d_lt_new = dataset_new();
#endif

	old_n_lt = dless->n;
#ifdef USE_BRUTE_FORCE
	n_gt = dataset_put_nans_at_end(dgreater, root, lenroot);
	n_lt = dataset_put_nans_at_end(dless, root, lenroot);
	dless->n = n_lt;
#endif
#ifdef USE_KDTREE
	// initialize the KDtree
	KDtree_init(&kdtree_less, dless, lenroot, root);
	KDtree_init(&kdtree_gt, dgreater, lenroot, root);
#endif /* USE_KDTREE */
#ifdef USE_RANGE_TREE
	n_gt = dataset_put_nans_at_end(dgreater, root, lenroot);
	range_tree_add_dim(rt, root[lenroot - 1], indices[root[lenroot-1]], dless->n);
#endif

	/* prune using upper bound */
	if(lenroot == 200000)
	{
	    u_bound = 0;
	    for(r = 0; r < whole_data_hrank->n; r++)
	    {
		double max = whole_data_hrank->data[r][root[0]];
	    
		for(t = 1; t < lenroot; t++)
		{
		    if(whole_data_hrank->data[r][root[t]] > max)
		    {
			max = whole_data_hrank->data[r][root[t]];
		    }
		}
		u_bound += N - max - 1;
	    }
	    for(r = 0; r < whole_data_lrank->n; r++)
	    {
		double min = whole_data_lrank->data[r][root[0]];
	    
		for(t = 1; t < lenroot; t++)
		{
		    if(whole_data_lrank->data[r][root[t]] < min)
		    {
			min = whole_data_lrank->data[r][root[t]];
		    }
		}
		u_bound += min;
	    }
	    u_bound /= 2;
	}
	/* compute lower bound */
	//l_bound = orig_data->n * (orig_data->n - 1) / 2;
	//for(r = 0; r < orig_data->n; r++)
	//{
	//    double max = orig_data->data[r][root[0]];
	//    double min = orig_data->data[r][root[0]];
	//
	//    for(t = 1; t < lenroot; t++)
	//    {
	//	if(orig_data->data[r][root[t]] > max)
	//	{
	//	    max = orig_data->data[r][root[t]];
	//	}
	//	if(orig_data->data[r][root[t]] < min)
	//	{
	//	    min = orig_data->data[r][root[t]];
	//	}
	//    }
	//    l_bound -= (max - min);
	//}
#endif /* SUPP_TAU */

#if defined (SUPP_RHO) || defined (SUPP_F)
	/* rho and F */
	F = 0;
	rho = 0;
	for(r = 0; r < orig_data->n; r++)
	{
	    double max = orig_data->data[r][root[0]];
	    double min = orig_data->data[r][root[0]];
	
	    for(t = 1; t < lenroot; t++)
	    {
		if(orig_data->data[r][root[t]] > max)
		{
		    max = orig_data->data[r][root[t]];
		}
		if(orig_data->data[r][root[t]] < min)
		{
		    min = orig_data->data[r][root[t]];
		}
	    }
	    F += (max - min);
	    rho += (max - min) * (max - min);
	}
#endif

#ifdef SUPP_TAU
	if(u_bound < minsup)
	{
	    if(lenroot > 2)
	    printf("pruned!!!, length %ld\n", lenroot);
	}
	else
#endif /* SUPP_TAU */
	{
#ifdef SUPP_TAU
	    /* compute support */
#ifdef USE_BRUTE_FORCE
	    supp = 0;
	    for(r = 0; r < n_gt; r++)
	    {
		size_t tmp;
		size_t tmp2;
	
		for(s = 0; s < lenroot; s++)
		    bounds[s] = dgreater->data[r][root[s]];
		tmp2 = 0;
		for(s = 0; s < n_lt; s++)
		{
		    size_t ok = 1;
		    for(t = 0; t < lenroot; t++)
			if(isnan(bounds[t]) || isnan(dless->data[s][root[t]]) || bounds[t] <= dless->data[s][root[t]])
			    ok = 0;
		    tmp2 += ok;
		}
		supp += tmp2;
		if(tmp2 > 0)
		    dataset_append_record(d_gt_new, dgreater->data[r]);
	    }
#endif /* USE_BRUTE_FORCE */

#ifdef USE_RANGE_TREE
	    supp = 0;
	    for(r = 0; r < n_gt; r++)
	    {
		size_t tmp;
	
		for(s = 0; s < lenroot; s++)
		    bounds[s] = dgreater->data[r][root[s]];
		tmp = range_tree_get_count(rt, bounds, lenroot);
		//printf("%d\n", tmp);
		supp += tmp;
		if(tmp > 0)
		    dataset_append_record(d_gt_new, dgreater->data[r]);
	    }
#endif /* USE_RANGE_TREE */

#ifdef USE_KDTREE
	    supp = KDtree_join(&kdtree_gt, &kdtree_less, mark_gt, mark_lt);
	    // remove unused records
	    for(u = 0; u < kdtree_gt.n; u++)
	    {
		if(mark_gt[u])
		    dataset_append_record(d_gt_new, kdtree_gt.data[u]);
	    }
	    for(u = 0; u < kdtree_less.n; u++)
	    {
		if(mark_lt[u])
		    dataset_append_record(d_lt_new, kdtree_less.data[u]);
	    }
	    //printf("%d,%d\n", d_lt_new->n, d_gt_new->n);
#endif /* USE_KDTREE */

#endif /* SUPP_TAU */
#ifdef SUPP_F
	    F = (double)N*(N-1) - F;
	    supp = F;
#endif
#ifdef SUPP_RHO
	    rho = (double)N*(N-1)*(N-1) - rho;
	    supp = rho / 1000000;
#endif


	    //printf("lenroot %d, supp=%d, u_bound=%d\n", lenroot,supp, u_bound);
#ifdef SUPP_TAU
	    if(supp > u_bound)
		printf("!!!!!!!!!! worng bound %ld, %ld!!!!!!!!!\n", supp, u_bound);
	    //if(supp < minsup && u_bound>= minsup)
	    //	printf("!!!!!!!!!! pruned by support %ld, %ld !!!!!!!!!\n", supp, u_bound);
#endif	     
	    //printf("===> %ld, [%ld, %dl] %f\n", supp, l_bound, u_bound, (double)(u_bound - l_bound) / l_bound);
	    //supp = u_bound;
	    if(supp >= minsup)
	    {
		//if(lenroot != 1)
		{
		    printf("[");
		    for(s = 0; s < lenroot; s++)
		    {
			printf("%ld", root[s]);
			if(s < lenroot - 1)
			    printf(", ");
		    }
		    //printf("] tau= %d ubound= %d F= %lf rho= %lf\n", supp, u_bound, F, rho);
		    printf("] %ld\n", supp);
		}
		// insert new frequent set into a hashtable
		{
		    size_t z;
		    size_t * key = malloc((lenroot + 1) * sizeof(size_t));
		    size_t * value = malloc(sizeof(size_t));

		    key[0] = lenroot;
		    for(z = 0; z < lenroot; z++)
		    {
			key[z+1] = root[z];
		    }
		    *value = supp;
		    htable_put(freq, key, value);
		}
	    }
	}
    }
    if(lenroot < 1 || supp >= minsup)
    {
	size_t j;
	for(j = n_items; j > i; j--)
	{
	    root[lenroot] = j-1;
#ifdef USE_RANGE_TREE
	    eclat(j, n_items, d_gt_new, d_lt_new, minsup, root, lenroot + 1, freq, rt, indices);
#else
	    eclat(j, n_items, d_gt_new, d_lt_new, minsup, root, lenroot + 1, freq, mark_gt, mark_lt);
#endif
	}
    }
#ifdef SUPP_TAU
    if(lenroot >= 1)
    {
#ifdef USE_KDTREE
	dataset_free(d_gt_new, 0);
	dataset_free(d_lt_new, 0);
#endif
#ifdef USE_BRUTE_FORCE
	dless->n = old_n_lt;
	dataset_free(d_gt_new, 0);
#endif
#ifdef USE_RANGE_TREE
	dataset_free(d_gt_new, 0);
#endif
    }
#ifdef USE_RANGE_TREE
    if(lenroot >= 1)
    {
	range_tree_remove_dimension(rt, lenroot+1);
    }
#endif
#endif /* SUPP_TAU */
}


static
size_t hashvect(void * v)
{
    size_t * vec = (size_t *)v;
    size_t hash = 0;
    size_t i;

    for(i = 0; i < vec[0]; i++)
	hash = (hash << 2) + vec[i+1];
    return hash;
}

static
int cmpvect(void * a, void * b)
{
    size_t * veca = (size_t *)a;
    size_t * vecb = (size_t *)b;
    size_t i;

    if(veca[0] != vecb[0])
	return 0;
    for(i = 0; i < veca[0]; i++)
	if(veca[i+1] != vecb[i+1])
	    return 0;
    return 1;
}

static
size_t hasstring(void * v)
{
    char * str = (char *)v;
    size_t hash = 0;
    size_t i;

    for(i = 0; str[i] != '\0'; i++)
	hash = (hash << 2) + str[i];
    return hash;
}

static
int cmpstring(void * a, void * b)
{
    char * veca = (char *)a;
    char * vecb = (char *)b;

    if(strcmp(a, b))
    {
	return 0;
    }
    return 1;
}

void
mine_freq(size_t n_items, dataset * dless, dataset * dgreater, size_t minsup)
{
#ifdef USE_RANGE_TREE
    index_entry ** indices = range_tree_create_indices(dless);
    range_tree * rt = range_tree_new(0.0);
#endif

    size_t root[5000];
    size_t i;
    htable * freq = htable_new(hashvect, cmpvect);
#ifdef USE_RANGE_TREE
    eclat(0, n_items, dgreater, dless, minsup, root, 0, freq, rt, indices);
#elif defined USE_KDTREE || defined USE_BRUTE_FORCE
    size_t mark_gt[dgreater->n];
    size_t mark_lt[dless->n];
    eclat(0, n_items, dgreater, dless, minsup, root, 0, freq, mark_gt, mark_lt);
#endif
#ifdef USE_EXPLICIT_PAIRS
    tidlist ** itdl = make_item_tidlists(dgreater, dless);
    eclat_items(0, NULL, n_items, itdl, minsup, root, 0, freq);
    for(i = 0; i < n_items; i++)
    {
	tidlist_free(itdl[i]);
    }
    free(itdl);
#endif
    htable_free(freq, 1, 1);


#ifdef USE_RANGE_TREE
    range_tree_free(rt);
    range_tree_free_indices(indices, n_items);
#endif
}

void
mine_freq_for_categ_attr(size_t * iset, int depth, 
			 htable ** categ_values,
			 dataset * categ_d_greater, dataset * d_greater,
			 dataset * categ_d_less, dataset * d_less,
			 size_t minsup)
{
    size_t c1 = 0, c2 = 1;
    size_t r, i;
    size_t max_supp;
    int last_attr;
    int attr_no;


    /* find all frequent in this class */
    if(depth > 0)
    mine_freq(d_less->n_items, d_less, d_greater, minsup);


    /* recursive call */
    if(depth == 0)
	last_attr = 0;
    else
	last_attr = iset[depth-1] + 1;

    //printf("%d,%d depth %d\n", last_attr, categ_d_less->n_items - 1, depth);
    for(attr_no = categ_d_less->n_items - 1; attr_no >= last_attr; attr_no--)
    {
	size_t n = categ_values[attr_no]->n;
	iset[depth] = attr_no;
	for(c1 = 0; c1 < n; c1++)
	{
	    for(c2 = 0; c2 < n; c2++)
	    {
		/* create databases for each pair attribute classes */
		dataset * dless_new = dataset_new();
		dataset * dgreater_new = dataset_new();
		dataset * categ_dless_new = dataset_new();
		dataset * categ_dgreater_new = dataset_new();
		dless_new->n_items = d_less->n_items;
		dgreater_new->n_items = d_greater->n_items;
		categ_dless_new->n_items = categ_d_less->n_items;
		categ_dgreater_new->n_items = categ_d_greater->n_items;
		for(r = 0; r < d_less->n; r++)
		{
		    if((int)(categ_d_less->data[r][attr_no]) == c1)
		    {
			dataset_append_record(dless_new, d_less->data[r]);
			dataset_append_record(categ_dless_new, categ_d_less->data[r]);
		    }
		}
		for(r = 0; r < d_greater->n; r++)
		{
		    if((int)(categ_d_greater->data[r][attr_no]) == c2)
		    {
			dataset_append_record(dgreater_new, d_greater->data[r]);
			dataset_append_record(categ_dgreater_new, categ_d_greater->data[r]);
		    }
		}
		printf("----------------- ");
		for(i = 0; i < depth+1; i++) printf("%ld ", iset[i]);
		printf(", classes %ld/%ld -------------------\n", c1, c2);
		max_supp = dless_new->n * dgreater_new->n;
		if(max_supp >= minsup)
		{
		    mine_freq_for_categ_attr(iset, depth + 1, categ_values,
					     categ_dgreater_new, dgreater_new,
					     categ_dless_new, dless_new,
					     minsup);
		    //mine_freq(d_less->n_items, dless_new, dgreater_new, minsup);
		}
		else
		{
		    printf("datasets too small\n");
		}
		dataset_free(dless_new, 0);
		dataset_free(dgreater_new, 0);
		dataset_free(categ_dless_new, 0);
		dataset_free(categ_dgreater_new, 0);
	    }
	}
    }
}

int
main(int argc, char ** argv)
{
    FILE * f;
    char * line;
    dataset * d = dataset_new();
    dataset * d_lrank;
    dataset * d_hrank;
    dataset * categ_d = dataset_new(); /* categorical columns of the dataset */

    size_t n_items;
    size_t n_categ_items;	/* assume categorical items come first */
    size_t minsup;

    size_t i, j;

    if(argc < 3)
    {
	fprintf(stderr, "usage: pairmining file minsupport\n");
	return 10;
    }
    minsup = atol(argv[2]);
    f = fopen(argv[1], "r");
    if(f == NULL)
    {
	fprintf(stderr, "error opening file\n");
	return 10;
    }


    /* read data characteristics */
    line = fgets(linebuf, MAXLINE, f);
    n_items = count_items(linebuf, &n_categ_items);
    n_items -= n_categ_items;
    d->n_items = n_items;
    categ_d->n_items = n_categ_items;
    //printf("%d numeric items/ %d categorical\n", n_items, n_categ_items);
    /* create lists of values for categorical attributes */
    htable ** categ_values = malloc(n_categ_items * sizeof(htable *));
    for(i = 0; i < n_categ_items; i++)
    {
	categ_values[i] = htable_new(hasstring, cmpstring);
    }

    /* read data into memory */
    while(line != NULL && strlen(line) > 0)
    {
	double * record;
	double * categ_record;
	char * str;
	char * str2;

	if(line[strlen(line) - 1] = '\n')
	{
	    line[strlen(line) - 1] = '\0';
	}
	categ_record = malloc(n_categ_items * sizeof(double));
	record = malloc(n_items * sizeof(double));
	str2 = line;
	str = strsep(&str2, ",;");
	j = 0;
	while(str != NULL)
	{
	    if(j < n_categ_items)
	    {
		if(str[0] == '\0' || !strcmp(str, "?"))
		{
		    categ_record[j] = NAN;
		}
		else
		{
		    void * cat_no = htable_get(categ_values[j], str);
		    if(cat_no == NULL)//add value to list
		    {
			cat_no = malloc(sizeof(size_t));
			*((size_t *)cat_no) = categ_values[j]->n;
			htable_put(categ_values[j], strdup(str), cat_no);
			//printf("%s = class %ld\n", str, *((size_t *)cat_no));
		    }
		    categ_record[j] = (float)(*(size_t *)cat_no);
		}
	    }
	    else
	    {
		if(str[0] == '\0' || !strcmp(str, "?"))
		{
		    record[j - n_categ_items] = NAN;
		}
		else
		{
		    record[j - n_categ_items] = atof(str);
		}
	    }
	    str = strsep(&str2, ",;");
	    j++;
	}
	/* append record to data */
	dataset_append_record(d, record);
	dataset_append_record(categ_d, categ_record);
	line = fgets(linebuf, MAXLINE, f);
    }
    fclose(f);
    N = d->n;			/* store initial database size */

    //print_data(d); 
    /* print_data(categ_d); */

    /* compute ranks */
    d_hrank = dataset_clone(d, 1);
    d_lrank = dataset_clone(d, 1);
    convert_to_ranks(d, TIES_AVG);
    convert_to_ranks(d_hrank, TIES_MAX);
    convert_to_ranks(d_lrank, TIES_MIN);
    orig_data = d;
    whole_data_hrank = d_hrank;
    whole_data_lrank = d_lrank;
    //printf("ranks\n");
    //print_data(d);
    
    if(n_categ_items == 0)
    {
	//printf("Only numerical attributes\n");
	mine_freq(d->n_items, d, d, minsup);
    }
    else
    {
	size_t iset[n_categ_items];
	mine_freq_for_categ_attr(iset, 0, categ_values, categ_d, d, categ_d, d, minsup);
    }

    /* free data */
    for(i = 0; i < n_categ_items; i++)
    {
	htable_free(categ_values[i], 1, 1);
    }
    free(categ_values);
    dataset_free(categ_d, 1);
    dataset_free(d_lrank, 1);
    dataset_free(d_hrank, 1);
    dataset_free(d, 1);

    return 0;
}
